""" .. Function(s) for the selection operator. """

from copy import deepcopy

import numpy as np


def selection(population, pop_fitness, best_prop, lucky_prop, maximise=False):
    """ Given a population, select a proportion of the "best"" individuals and
    another of the "lucky" individuals (if they are available) to form a set of
    potential parents.

    Parameters
    ----------
    population : list
        All current individuals.
    pop_fitness : list
        The fitness of each individual in :code:`population`.
    best_prop : float
        The proportion of the fittest individuals in :code:`population` to be
        selected.
    lucky_prop : float
        The proportion of lucky individuals in :code:`population` to be
        selected.
    maximise : bool, optional
        Determines whether an individual's fitness should be maximal or not.
        Defaults to :code:`False`.

    Returns
    -------
    parents : list
        The individuals chosen to potentially become parents.

    Raises
    ------
    ValueError
        If :code:`int(best_prop * len(population)) == 0` and
        :code:`int(lucky_prop * len(population)) == 0`.
    """

    size = len(population)
    num_best = int(best_prop * size)
    num_lucky = int(lucky_prop * size)

    if maximise:
        best_choice = np.argmax
    else:
        best_choice = np.argmin

    if num_best == 0 and num_lucky == 0:
        raise ValueError(
            'Not a large enough proportion of "best" and/or \
                          "lucky" individuals chosen. Reconsider these values.'
        )

    population = deepcopy(population)
    pop_fitness = deepcopy(pop_fitness)
    parents = []
    for _ in range(num_best):
        if population != []:
            best = best_choice(pop_fitness)
            pop_fitness.pop(best)
            parents.append(population.pop(best))

    for _ in range(num_lucky):
        if population != []:
            lucky = np.random.choice(len(population))
            parents.append(population.pop(lucky))

    return parents



""" All currently implemented, discrete distribution classes. """

import numpy as np

from .base import Distribution


class Bernoulli(Distribution):
    """ Discrete column class given by the Bernoulli distribution. That is, a
    binomial distribution with exactly one trial.

    Attributes
    ----------
    name : str
        Name of the distribution, :code:`"Bernoulli"`.
    dtype : str
        Preferred datatype of distribution, :code:`"int"`.
    param_limits : dict
        A dictionary of the limits on the distribution parameter. Defaults to
        :code:`[0, 1]` for :code:`prob`.
    prob : float
        The success probability, sampled from :code:`param_limits["prob"]`.
        `Instance attribute.`
    """

    name = "Bernoulli"
    dtype = "int"
    hard_limits = {"prob": [0, 1]}
    param_limits = {"prob": [0, 1]}

    def __init__(self):

        self.prob = np.random.uniform(*self.param_limits["prob"])

    def sample(self, nrows):
        """ Take a sample of size :code:`nrows` from the Bernoulli distribution
        with success probability :code:`prob`. """

        return np.random.binomial(n=1, p=self.prob, size=nrows)


class Poisson(Distribution):
    """ Discrete column class given by the Poisson distribution.

    Attributes
    ----------
    name : str
        Name of distribution, :code:`"Poisson"`.
    dtype : str
        Preferred datatype of distribution, :code:`"int"`.
    param_limits : dict
        A dictionary of the limits of the distribution parameter. Defaults to
        :code:`[0, 10]` for :code:`lam`.
    lam : float
        The rate parameter, sampled from :code:`param_limits["lam"]`. `Instance
        attribute.`
    """

    name = "Poisson"
    dtype = "int"
    hard_limits = {"lam": [0, 10]}
    param_limits = {"lam": [0, 10]}

    def __init__(self):

        self.lam = np.random.uniform(*self.param_limits["lam"])

    def sample(self, nrows):
        """ Take a sample of size :code:`nrows` from the Poisson distribution
        with rate parameter :code:`lam`. """

        return np.random.poisson(lam=self.lam, size=nrows)



""" A collection of useful functions for various processes within the GA. """

import numpy as np


def _get_pdf_counts(metadata, pdfs):
    """ Get the count of each pdf class present in metadata. """

    return {
        pdf_class: sum([isinstance(pdf, pdf_class) for pdf in metadata])
        for pdf_class in pdfs
    }


def _rename(dataframe, axis):
    """ Rename metadata or reindex to make sense after deletion or addition of a
    new line. """

    if axis == 0:
        dataframe = dataframe.reset_index(drop=True)
    else:
        dataframe.columns = range(len(dataframe.columns))

    return dataframe


def _fillna(dataframe, metadata):
    """ Fill in `NaN` values of a column by sampling from the distribution
    associated with it. """

    for col, pdf in zip(dataframe.columns, metadata):
        data = dataframe[col]
        if data.isnull().any():
            nulls = data.isnull()
            samples = pdf.sample(nulls.sum())
            dataframe.loc[nulls, col] = samples

        dataframe[col] = dataframe[col].astype(pdf.dtype)

    return dataframe


def _remove_line(dataframe, metadata, axis, col_limits=None, pdfs=None):
    """ Remove a line (row or column) from a dataset at random. """

    if axis == 0:
        line = np.random.choice(dataframe.index)
        dataframe = _rename(dataframe.drop(line, axis=axis), axis)

    elif isinstance(col_limits[0], tuple):
        ncols = dataframe.shape[1]
        pdf_counts = _get_pdf_counts(metadata, pdfs)
        while len(dataframe.columns) != ncols - 1:
            line = np.random.choice(dataframe.columns)
            column_idx = dataframe.columns.get_loc(line)
            pdf = metadata[column_idx]
            pdf_class = pdf.__class__
            pdf_idx = pdfs.index(pdf_class)
            if pdf_counts[pdf_class] > col_limits[0][pdf_idx]:
                dataframe = _rename(dataframe.drop(line, axis=axis), axis)
                metadata.pop(column_idx)

    else:
        line = np.random.choice(dataframe.columns)
        idx = dataframe.columns.get_loc(line)
        dataframe = _rename(dataframe.drop(line, axis=axis), axis)
        metadata.pop(idx)

    return dataframe, metadata


def _add_line(
    dataframe, metadata, axis, col_limits=None, pdfs=None, weights=None
):
    """ Add a line (row or column) to the end of a dataset. Rows are added by
    sampling from the distribution associated with that column in `metadata`.
    metadata are added in the same way that they are at the initial creation of
    an individual by sampling from the list of all `pdfs` according to
    `weights`. """

    nrows, ncols = dataframe.shape

    if axis == 0:
        dataframe = dataframe.append(
            {i: pdf.sample(1)[0] for i, pdf in enumerate(metadata)},
            ignore_index=True,
        )

    elif isinstance(col_limits[1], tuple):
        pdf_counts = _get_pdf_counts(metadata, pdfs)
        while len(dataframe.columns) != ncols + 1:
            pdf_class = np.random.choice(pdfs, p=weights)
            idx = pdfs.index(pdf_class)
            pdf = pdf_class()
            if pdf_counts[pdf_class] < col_limits[1][idx]:
                dataframe[ncols + 1] = pdf.sample(nrows)
                metadata.append(pdf)
    else:
        pdf_class = np.random.choice(pdfs, p=weights)
        pdf = pdf_class()
        dataframe[ncols + 1] = pdf.sample(nrows)
        metadata.append(pdf)

    dataframe = _rename(dataframe, axis)
    return dataframe, metadata



""" Column pdf class imports. """

# pylint: disable=invalid-name

from .base import Distribution
from .continuous import Gamma, Normal
from .discrete import Bernoulli, Poisson


all_pdfs = [Bernoulli, Gamma, Normal, Poisson]

continuous_pdfs = [pdf for pdf in all_pdfs if pdf.dtype == "float"]
discrete_pdfs = [pdf for pdf in all_pdfs if pdf.dtype == "int"]



""" Function for compacting the search space. """


def _get_all_param_values(parents):
    """ Get all the parent parameter values present in a dictionary. """

    all_param_vals = {}
    for _, metadata in parents:
        for column in metadata:
            col = column.__class__
            if col not in all_param_vals.keys():
                all_param_vals[col] = {}
            for param_name, param_val in column.__dict__.items():
                try:
                    all_param_vals[col][param_name] += [param_val]
                except KeyError:
                    all_param_vals[col][param_name] = [param_val]

    return all_param_vals


def compact_search_space(parents, pdfs, itr, max_iter, compaction_ratio):
    """ Given the current progress of the GA, compact the search space, i.e. the
    parameter spaces for each of the distribution classes in :code:`pdfs`. """

    if (
        compaction_ratio in [0, 1]
        or compaction_ratio < 0
        or compaction_ratio > 1
    ):
        raise ValueError("Compaction ratio, s, must satisfy 0 < s < 1.")

    compact_factor = 1 - (itr / (compaction_ratio * max_iter))
    all_param_vals = _get_all_param_values(parents)

    for pdf, params in all_param_vals.items():
        for name, vals in params.items():

            midpoint = sum(vals) / len(vals)
            shift = (max(vals) - min(vals)) * compact_factor / 2

            lower = midpoint - shift
            upper = midpoint + shift

            if shift < 0:
                lower, upper = upper, lower

            lower = max([pdf.hard_limits[name][0], lower])
            upper = min([pdf.hard_limits[name][1], upper])

            pdf.param_limits[name] = [lower, upper]

    return pdfs



""" Setup script. """

# pylint: disable=exec-used,undefined-variable

from setuptools import find_packages, setup


with open("README.rst", "r") as readme_file:
    README = readme_file.read()

exec(open("src/edo/version.py", "r").read())

setup(
    name="edo",
    version=__version__,
    description="Generating artificial datasets through genetic evolution.",
    long_description=README,
    long_description_content_type="text/x-rst",
    url="https://github.com/daffidwilde/edo",
    author="Henry Wilde",
    author_email="henrydavidwilde@gmail.com",
    license="MIT",
    keywords=["genetic-algorithm" "data" "evolution"],
    packages=find_packages("src"),
    package_dir={"": "src"},
)



""" .. The main script containing a generic genetic algorithm. """

import numpy as np

from .compact import compact_search_space
from .fitness import get_fitness
from .operators import selection
from .population import create_initial_population, create_new_population


def run_algorithm(
    fitness,
    size,
    row_limits,
    col_limits,
    pdfs,
    weights=None,
    stop=None,
    dwindle=None,
    max_iter=100,
    best_prop=0.25,
    lucky_prop=0,
    crossover_prob=0.5,
    mutation_prob=0.01,
    compact=None,
    maximise=False,
    seed=None,
    fitness_kwargs=None,
):
    """ Run a genetic algorithm under the presented constraints, giving a
    population of artificial datasets for which the given fitness function
    performs well.

    Parameters
    ----------
    fitness : func
        Any real-valued function that takes one :class:`pandas.DataFrame` as
        argument. Any further arguments should be passed to
        :code:`fitness_kwargs`.
    size : int
        The size of the population to create.
    row_limits : list
        Lower and upper bounds on the number of rows a dataset can have.
    col_limits : list
        Lower and upper bounds on the number of columns a dataset can have.

        Tuples can also be used to specify the min/maximum number of columns
        there can be of each type in :code:`pdfs`.
    pdfs : list
        Used to create the initial population and instruct the GA how a column
        should be manipulated in a dataset.

        .. note::
            For reproducibility, a user-defined class' :code:`sample` method
            should use NumPy for any random elements as the seed for the GA is
            set using :func:`np.random.seed`.
    weights : list
        A probability distribution on how to select columns from
        :code:`pdfs`. If :code:`None`, pdfs will be chosen uniformly.
    stop : func
        A function which acts as a stopping condition on the GA. Such functions
        should take only the fitness of the current population as argument, and
        should return a boolean variable. If :code:`None`, the GA will run up
        until its maximum number of iterations.
    dwindle : func
        A function which acts as a means of dwindling the mutation probability.
        Such functions should take the current mutation probability and the
        current iteration as argument, and should return a new mutation
        probability. If :code:`None`, the GA will run with a constant mutation
        probability.
    max_iter : int
        The maximum number of iterations to be carried out before terminating.
    best_prop : float
        The proportion of a population from which to select the "best" potential
        parents.
    lucky_prop : float
        The proportion of a population from which to sample some "lucky"
        potential parents. Set to zero as standard.
    crossover_prob : float
        The probability with which to sample dimensions from the first parent
        over the second in a crossover operation. Defaults to 0.5.
    mutation_prob : float
        The probability of a particular characteristic in an individual's
        dataset being mutated. If using :code:`dwindle`, this is an initial
        probability.
    compact : float
        The proportion of the maximum iterations to reduce and focus the
        mutation search space. Defaults to `None` but must be between 0 and 1
        (not inclusive).
    maximise : bool
        Determines whether :code:`fitness` is a function to be maximised or not.
        Fitness scores are minimised by default.
    seed : int
        The seed for a particular run of the genetic algorithm. If :code:`None`,
        no seed is set.
    fitness_kwargs : dict
        Any additional parameters that need to be passed to :code:`fitness` should
        be placed here as a dictionary or suitable mapping.

    Returns
    -------
    population : list
        The final population.
    pop_fitness : list
        The fitness of all individuals in the final population.
    all_populations : list
        Every population in each generation.
    all_fitnesses : list
        Every individual's fitness in each generation.
    """

    if seed is not None:
        np.random.seed(seed)

    population = create_initial_population(
        size, row_limits, col_limits, pdfs, weights
    )

    pop_fitness = get_fitness(fitness, population, fitness_kwargs)

    converged = False
    if stop:
        converged = stop(pop_fitness)

    itr = 0
    all_populations, all_fitnesses = [population], [pop_fitness]
    while itr < max_iter and not converged:

        parents = selection(
            population, pop_fitness, best_prop, lucky_prop, maximise
        )

        population = create_new_population(
            parents,
            size,
            crossover_prob,
            mutation_prob,
            row_limits,
            col_limits,
            pdfs,
            weights,
        )

        pop_fitness = get_fitness(fitness, population, fitness_kwargs)

        all_populations.append(population)
        all_fitnesses.append(pop_fitness)

        if stop:
            converged = stop(pop_fitness)
        if dwindle:
            mutation_prob = dwindle(mutation_prob, itr)
        if compact is not None:
            pdfs = compact_search_space(parents, pdfs, itr, max_iter, compact)
        itr += 1

    return population, pop_fitness, all_populations, all_fitnesses
